# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add-entrypoint-to-container

Revision ID: f979327df44b
Revises: 74c97dca93d0
Create Date: 2020-03-30 03:19:01.080873

"""

# revision identifiers, used by Alembic.
revision = 'f979327df44b'
down_revision = '74c97dca93d0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

import zun


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('container', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                'entrypoint',
                zun.db.sqlalchemy.models.JSONEncodedList(),
                nullable=True))

    # ### end Alembic commands ###
